/*
 * SpanDSP - a series of DSP components for telephony
 *
 * t31.c - A T.31 compatible class 1 FAX modem interface.
 *
 * Written by Steve Underwood <steveu@coppice.org>
 *
 * Copyright (C) 2004 Steve Underwood
 *
 * All rights reserved.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 * $Id: t31.c,v 1.4 2004/10/04 05:57:02 steveu Exp $
 */

/*! \file */

#include <stdint.h>
#include <stdlib.h>
#include <unistd.h>
#include <stdio.h>
#include <fcntl.h>
#include <string.h>
#include <math.h>
#include <assert.h>

#include <tiffio.h>

#include "spandsp.h"

t31_profile_t profiles[3];

static char *manufacturer = "www.opencall.org";
static char *model = "spandsp";
static char *revision = "1.0";

enum
{
    ASCII_RESULT_CODES = 1,
    NUMERIC_RESULT_CODES,
    NO_RESULT_CODES
};

#define DLE 0x10
#define ETX 0x03
#define SUB 0x1A

enum
{
    T31_SILENCE_TX,
    T31_CED_TONE,
    T31_CNG_TONE,
    T31_V21_TX,
    T31_V17_TX,
    T31_V27TER_TX,
    T31_V29_TX,
    T31_V21_RX,
    T31_V17_RX,
    T31_V27TER_RX,
    T31_V29_RX
};

enum
{
    RESPONSE_CODE_OK = 0,
    RESPONSE_CODE_CONNECT,
    RESPONSE_CODE_RING,
    RESPONSE_CODE_NO_CARRIER,
    RESPONSE_CODE_ERROR,
    RESPONSE_CODE_XXX,
    RESPONSE_CODE_NO_DIALTONE,
    RESPONSE_CODE_BUSY,
    RESPONSE_CODE_NO_ANSWER,
    RESPONSE_CODE_FCERROR
};

const char *response_codes[] =
{
    "OK",
    "CONNECT",
    "RING",
    "NO CARRIER",
    "ERROR",
    "???",
    "NO DIALTONE",
    "BUSY",
    "NO ANSWER",
    "+FCERROR",
};

static void at_put_string(t31_state_t *s, const char *t)
{
    /* Ensure we do not overflow the buffer, but do nothing to prevent loss of
       data. */
    while (*t  &&  s->rx_data_bytes < sizeof(s->rx_data))
        s->rx_data[s->rx_data_bytes++] = *t++;
}
/*- End of function --------------------------------------------------------*/

static void at_put_response(t31_state_t *s, const char *t)
{
    /* Ensure we do not overflow the buffer, but do nothing to prevent loss of
       data. */
    if (s->verbose == ASCII_RESULT_CODES  &&  s->rx_data_bytes < sizeof(s->rx_data) - 1)
    {
        s->rx_data[s->rx_data_bytes++] = '\r';
        s->rx_data[s->rx_data_bytes++] = '\n';
    }
    while (*t  &&  s->rx_data_bytes < sizeof(s->rx_data))
        s->rx_data[s->rx_data_bytes++] = *t++;
    if (s->rx_data_bytes < sizeof(s->rx_data) - 1)
    {
        s->rx_data[s->rx_data_bytes++] = '\r';
        s->rx_data[s->rx_data_bytes++] = '\n';
    }
}
/*- End of function --------------------------------------------------------*/

static void at_put_numeric_response(t31_state_t *s, int val)
{
    char buf[20];

    snprintf(buf, sizeof(buf), "%d", val);
    at_put_response(s, buf);
}
/*- End of function --------------------------------------------------------*/

static void at_put_3digit_numeric_response(t31_state_t *s, int val)
{
    char buf[4];

    snprintf(buf, sizeof(buf), "%3.3d", val);
    at_put_response(s, buf);
}
/*- End of function --------------------------------------------------------*/

static void at_put_response_code(t31_state_t *s, int code)
{
    char buf[20];

    switch (s->verbose)
    {
    case ASCII_RESULT_CODES:
        at_put_response(s, response_codes[code]);
        break;
    case NUMERIC_RESULT_CODES:
        snprintf(buf, sizeof(buf), "%d\r", code);
        at_put_string(s, buf);
        break;
    default:
        /* No result codes */
        break;
    }
}
/*- End of function --------------------------------------------------------*/

static int restart_modem(t31_state_t *s, int new_modem, int new_transmit)
{
    tone_gen_descriptor_t tone_desc;

    if (s->modem == new_modem)
        return 0;
    queue_flush(&(s->rx_queue));
    s->modem = new_modem;
    switch (s->modem)
    {
    case T31_CED_TONE:
        s->silent_samples += SAMPLE_RATE*0.2;
        make_tone_gen_descriptor(&tone_desc,
                                 2100,
                                 -11,
                                 0,
                                 0,
                                 2600,
                                 75,
                                 0,
                                 0,
                                 FALSE);
        tone_gen_init(&(s->tone_gen), &tone_desc);
        break;
    case T31_CNG_TONE:
        /* 0.5s of 1100Hz + 3.0s of silence repeating */
        make_tone_gen_descriptor(&tone_desc,
                                 1100,
                                 -11,
                                 0,
                                 0,
                                 500,
                                 3000,
                                 0,
                                 0,
                                 TRUE);
        tone_gen_init(&(s->tone_gen), &tone_desc);
        break;
    case T31_V21_TX:
        fsk_tx_init(&(s->v21tx), &preset_fsk_specs[FSK_V21CH2], (get_bit_func_t) hdlc_tx_getbit, &(s->hdlctx));
        break;
    case T31_V21_RX:
        fsk_rx_init(&(s->v21rx), &preset_fsk_specs[FSK_V21CH2], TRUE, (put_bit_func_t) hdlc_rx_bit, &(s->hdlcrx));
        break;
#if defined(ENABLE_V17)
    case T31_V17_TX:
        v17_tx_restart(&(s->v17tx), s->bit_rate, s->short_train);
        break;
    case T31_V17_RX:
        v17_rx_restart(&(s->v17rx), s->bit_rate, s->short_train);
        break;
#endif
    case T31_V27TER_TX:
        v27ter_tx_restart(&(s->v27ter_tx), s->bit_rate);
        break;
    case T31_V27TER_RX:
        v27ter_rx_restart(&(s->v27ter_rx), s->bit_rate);
        break;
    case T31_V29_TX:
        v29_tx_restart(&(s->v29tx), s->bit_rate);
        break;
    case T31_V29_RX:
        v29_rx_restart(&(s->v29rx), s->bit_rate);
        break;
    }
    s->bit_no = 0;
    s->current_byte = 0xFF;
    s->tx_in_bytes = 0;
    s->transmit = new_transmit;
    return 0;
}
/*- End of function --------------------------------------------------------*/

static void fast_putbit(void *user_data, int bit)
{
    t31_state_t *s;

    s = (t31_state_t *) user_data;
    if (bit < 0)
    {
        /* Special conditions */
        switch (bit)
        {
        case PUTBIT_TRAINING_FAILED:
            break;
        case PUTBIT_TRAINING_SUCCEEDED:
            /* The modem is now trained */
            at_put_response_code(s, RESPONSE_CODE_CONNECT);
            s->rx_signal_present = TRUE;
            break;
        case PUTBIT_CARRIER_UP:
            break;
        case PUTBIT_CARRIER_DOWN:
            if (s->rx_signal_present)
            {
                s->rx_data[s->rx_data_bytes++] = DLE;
                s->rx_data[s->rx_data_bytes++] = ETX;
                at_put_response_code(s, RESPONSE_CODE_NO_CARRIER);
                s->state = AT_MODE_COMMAND;
            }
            s->rx_signal_present = FALSE;
            break;
        default:
            if (s->verbose)
                fprintf(stderr, "Eh!\n");
            break;
        }
        return;
    }
    s->current_byte = (s->current_byte >> 1) | (bit << 7);
    if (++s->bit_no >= 8)
    {
        if (s->current_byte == DLE)
            s->rx_data[s->rx_data_bytes++] = s->current_byte;
        s->rx_data[s->rx_data_bytes++] = s->current_byte;
        s->bit_no = 0;
        s->current_byte = 0;
    }
}
/*- End of function --------------------------------------------------------*/

static int fast_getbit(void *user_data)
{
    t31_state_t *s;
    int bit;

    s = (t31_state_t *) user_data;
    bit = s->current_byte & 1;
    s->current_byte >>= 1;
    if (++s->bit_no >= 8)
    {
        s->bit_no = 0;
        if (s->tx_data_bytes < s->tx_in_bytes)
            s->current_byte = s->tx_data[s->tx_data_bytes++];
        else
            s->current_byte = 0xFF;
    }
    return bit;
}
/*- End of function --------------------------------------------------------*/

static void hdlc_tx_underflow(void *user_data)
{
    t31_state_t *s;

    s = (t31_state_t *) user_data;
    if (s->hdlc_final)
    {
        at_put_response_code(s, RESPONSE_CODE_OK);
        s->state = AT_MODE_COMMAND;
        s->hdlc_final = FALSE;
    }
    else
    {
        at_put_response_code(s, RESPONSE_CODE_CONNECT);
    }
}
/*- End of function --------------------------------------------------------*/

static void hdlc_accept(void *user_data, int ok, const uint8_t *msg, int len)
{
    char buf[256];
    t31_state_t *s;
    int i;

    s = (t31_state_t *) user_data;
    if (len < 0)
    {
        /* Special conditions */
        switch (len)
        {
        case PUTBIT_CARRIER_UP:
            s->rx_signal_present = TRUE;
            break;
        case PUTBIT_CARRIER_DOWN:
            if (s->dte_is_waiting)
            {
                at_put_response_code(s, RESPONSE_CODE_NO_CARRIER);
            }
            else
            {
                buf[0] = RESPONSE_CODE_NO_CARRIER;
                queue_write_msg(&(s->rx_queue), buf, 1);
            }
            s->rx_signal_present = FALSE;
            break;
        case PUTBIT_FRAMING_OK:
            if (s->modem == T31_CNG_TONE)
            {
                /* Once we get any valid HDLC, the CNG tone should
                   stop. */
                s->modem = T31_CNG_TONE;
                s->transmit = FALSE;
            }
            break;
        default:
            fprintf(stderr, "Unexpected HDLC special length - %d!\n", len);
            break;
        }
        return;
    }
    
    if (s->dte_is_waiting)
    {
fprintf(stderr, "dte is waiting - %d\n", len);
        at_put_response_code(s, RESPONSE_CODE_CONNECT);
        /* Send straight away */
        for (i = 0;  i < len;  i++)
        {
            if (msg[i] == DLE)
                s->rx_data[s->rx_data_bytes++] = DLE;
            s->rx_data[s->rx_data_bytes++] = msg[i];
        }
        s->rx_data[s->rx_data_bytes++] = 0;
        s->rx_data[s->rx_data_bytes++] = 0;
        s->rx_data[s->rx_data_bytes++] = DLE;
        s->rx_data[s->rx_data_bytes++] = ETX;
        if (ok)
            at_put_response_code(s, RESPONSE_CODE_OK);
        else
            at_put_response_code(s, RESPONSE_CODE_ERROR);
        s->dte_is_waiting = FALSE;
        s->state = AT_MODE_COMMAND;
    }
    else
    {
fprintf(stderr, "queue it - %d\n", len);
        /* Queue it */
        buf[0] = (ok)  ?  RESPONSE_CODE_OK  :  RESPONSE_CODE_FCERROR;
        memcpy(buf + 1, msg, len);
        queue_write_msg(&(s->rx_queue), buf, len + 1);
    }
}
/*- End of function --------------------------------------------------------*/

static void dle_unstuff_hdlc(t31_state_t *s, const char *stuffed, int len)
{
    int i;

    for (i = 0;  i < len;  i++)
    {
        if (s->dled)
        {
            s->dled = FALSE;
            if (stuffed[i] == ETX)
            {
                hdlc_tx_preamble(&(s->hdlctx), 2);
                hdlc_tx_frame(&(s->hdlctx), s->hdlc_buf, s->hdlc_len);
                hdlc_tx_preamble(&(s->hdlctx), 2);
                s->hdlc_final = (s->hdlc_buf[1] & 0x10);
                s->hdlc_len = 0;
            }
            else if (stuffed[i] == SUB)
            {
                s->hdlc_buf[s->hdlc_len++] = DLE;
                s->hdlc_buf[s->hdlc_len++] = DLE;
            }
            else
            {
                s->hdlc_buf[s->hdlc_len++] = stuffed[i];
            }
        }
        else
        {
            if (stuffed[i] == DLE)
                s->dled = TRUE;
            else
                s->hdlc_buf[s->hdlc_len++] = stuffed[i];
        }
    }
}
/*- End of function --------------------------------------------------------*/

static void dle_unstuff(t31_state_t *s, const char *stuffed, int len)
{
    int i;
    
    for (i = 0;  i < len;  i++)
    {
        if (s->dled)
        {
            s->dled = FALSE;
            if (stuffed[i] == ETX)
            {
                fprintf(stderr, "%d byte data\n", s->hdlc_len);
                if ((s->hdlc_buf[1] & 0x10))
                {
                    s->state = AT_MODE_COMMAND;
                    at_put_response_code(s, RESPONSE_CODE_OK);
                }
                else
                {
                    at_put_response_code(s, RESPONSE_CODE_CONNECT);
                }
                s->hdlc_len = 0;
            }
        }
        else
        {
            if (stuffed[i] == DLE)
            {
                s->dled = TRUE;
            }
            else
            {
                s->tx_data[s->tx_in_bytes++];
            }
        }
    }
}
/*- End of function --------------------------------------------------------*/

static int at_get_reg(t31_state_t *s, int reg, uint8_t *val)
{
    *val = s->s_regs[reg];
    return TRUE;
}
/*- End of function --------------------------------------------------------*/

static int at_set_reg(t31_state_t *s, int reg, uint8_t val)
{
    s->s_regs[reg] = val;
    return TRUE;
}
/*- End of function --------------------------------------------------------*/

static int at_set_bit(t31_state_t *s, int reg, int bit, int val)
{
    if (val)
        s->s_regs[reg] |= (1 << bit);
    else
        s->s_regs[reg] &= ~(1 << bit);
    return TRUE;
}
/*- End of function --------------------------------------------------------*/

static int at_get_bit(t31_state_t *s, int reg, int bit, int *val)
{
    *val = (s->s_regs[reg] >> bit) & 1;
    return TRUE;
}
/*- End of function --------------------------------------------------------*/

static int at_set_bits(t31_state_t *s, int x, int y, int z, uint8_t val)
{
    return TRUE;
}
/*- End of function --------------------------------------------------------*/

static int at_cid_mode(t31_state_t *s, int mode)
{
    return 10;
}
/*- End of function --------------------------------------------------------*/

static char *at_modem_class(t31_state_t *s, char *mode)
{
    return "1";
}
/*- End of function --------------------------------------------------------*/

static int at_dial_out(t31_state_t *s, const char *t)
{
    fprintf(stderr, "dialing '%s'\n", t);
    return 1;
}
/*- End of function --------------------------------------------------------*/

static int parse_num(const char **s, int max_value)
{
    int i;
    
    /* The spec says no digits is valid, and should be treated as zero. */
    i = 0;
    while (isdigit(**s))
    {
        i = i*10 + ((**s) & 0x0F);
        (*s)++;
    }
    if (i > max_value)
        i = -1;
    return i;
}
/*- End of function --------------------------------------------------------*/

static int parse_out(t31_state_t *s, const char **t, int *target, int max_value, const char *def)
{
    char buf[20];
    int val;

    switch (*(*t)++)
    {
    case '=':
        switch (**t)
        {
        case '?':
            /* Show possible values */
            (*t)++;
            at_put_response(s, def);
            break;
        default:
            /* Set value */
            if ((val = parse_num(t, max_value)) >= 0)
            {
                if (target)
                    *target = val;
            }
            else
            {
                return FALSE;
            }
            break;
        }
        break;
    case '?':
        /* Show current value */
        if (target)
        {
            snprintf(buf, sizeof(buf), "%d", *target);
            at_put_response(s, buf);
        }
        else
        {
            at_put_response(s, "0");
        }
        break;
    default:
        return FALSE;
    }
    return TRUE;
}
/*- End of function --------------------------------------------------------*/

static int process_class1_cmd(t31_state_t *s, const char **cmd)
{
    int val;
    int operation;
    int new_modem;
    int new_transmit;
    int i;
    int len;
    char *allowed;
    uint8_t msg[256];

    /* All class 1 FAX commands are supposed to give an ERROR response, if the phone
       is on-hook. */
    if (!s->connection_established)
    {
        at_put_response_code(s, RESPONSE_CODE_ERROR);
        return TRUE;
    }
    
    new_transmit = (*(*cmd - 2) == 'T');
    operation = *(*cmd - 1);

    val = -1;
    switch (operation)
    {
    case 'S':
        allowed = "0-255";
        break;
    case 'H':
        allowed = "3";
        break;
    default:
#if defined(ENABLE_V17)
        allowed = "24,48,72,96";
#else
        allowed = "24,48,72,73,74,96,97,98,121,122,145,146";
#endif
        break;
    }
    
    if (parse_out(s, cmd, &val, 255, allowed))
    {
        if (val >= 0)
        {
            if (!s->connection_established)
            {
fprintf(stderr, "No connection\n");
                return FALSE;
            }
            switch (operation)
            {
            case 'S':
                s->transmit = new_transmit;
                s->modem = T31_SILENCE_TX;
                if (new_transmit)
                {
                    s->silent_samples += val*80;
                }
                else
                {
                    queue_flush(&(s->rx_queue));
                    at_put_response_code(s, RESPONSE_CODE_OK);
                }
                fprintf(stderr, "Silence %dms\n", val*10);
                break;
            case 'H':
                switch (val)
                {
                case 3:
                    new_modem = (new_transmit)  ?  T31_V21_TX  :  T31_V21_RX;
                    s->short_train = FALSE;
                    s->bit_rate = 300;
                    break;
                default:
                    return FALSE;
                }
                break;
                fprintf(stderr, "HDLC\n");
                if (new_transmit != s->transmit)
                {
                    if (new_transmit)
                    {
fprintf(stderr, "New tx HDLC\n");
                        s->hdlc_final = FALSE;
                        hdlc_tx_init(&(s->hdlctx), hdlc_tx_underflow, s);
                        hdlc_tx_preamble(&(s->hdlctx), 40);
                    }
                    else
                    {
                        hdlc_rx_init(&(s->hdlcrx), hdlc_accept, s);
                        hdlc_rx_bad_frame_control(&(s->hdlcrx), TRUE);
                    }
                    s->hdlc_len = 0;
                    s->dled = FALSE;
                }
                if (new_transmit)
                {
                    at_put_response_code(s, RESPONSE_CODE_CONNECT);
                    s->state = AT_MODE_HDLC;
                }
                else
                {
                    /* Send straight away */
                    do
                    {
                        if (!queue_empty(&(s->rx_queue)))
                        {
                            len = queue_read_msg(&(s->rx_queue), msg, 256);
fprintf(stderr, "send now - %d %x %x %x %x\n", len, msg[0], msg[1], msg[2], msg[3]);
                            if (len > 1)
                            {
                                if (msg[0] == RESPONSE_CODE_OK)
                                    at_put_response_code(s, RESPONSE_CODE_CONNECT);
                                for (i = 1;  i < len;  i++)
                                {
                                    if (msg[i] == DLE)
                                        s->rx_data[s->rx_data_bytes++] = DLE;
                                    s->rx_data[s->rx_data_bytes++] = msg[i];
                                }
                                s->rx_data[s->rx_data_bytes++] = 0;
                                s->rx_data[s->rx_data_bytes++] = 0;
                                s->rx_data[s->rx_data_bytes++] = DLE;
                                s->rx_data[s->rx_data_bytes++] = ETX;
                            }
                            at_put_response_code(s, msg[0]);
                            if (msg[0] != RESPONSE_CODE_CONNECT)
                            {
                                s->dte_is_waiting = FALSE;
                                s->state = AT_MODE_COMMAND;
                            }
                        }
                        else
                        {
fprintf(stderr, "dte now waits\n");
                            s->dte_is_waiting = TRUE;
                            break;
                        }
                    }
                    while (msg[0] == RESPONSE_CODE_CONNECT);
                }
                restart_modem(s, new_modem, new_transmit);
                break;
            default:
                switch (val)
                {
                case 24:
                    new_modem = (new_transmit)  ?  T31_V27TER_TX  :  T31_V27TER_RX;
                    s->short_train = FALSE;
                    s->bit_rate = 2400;
                    break;
                case 48:
                    new_modem = (new_transmit)  ?  T31_V27TER_TX  :  T31_V27TER_RX;
                    s->short_train = FALSE;
                    s->bit_rate = 4800;
                    break;
                case 72:
                    new_modem = (new_transmit)  ?  T31_V29_TX  :  T31_V29_RX;
                    s->short_train = FALSE;
                    s->bit_rate = 7200;
                    break;
                case 96:
                    new_modem = (new_transmit)  ?  T31_V29_TX  :  T31_V29_RX;
                    s->short_train = FALSE;
                    s->bit_rate = 9600;
                    break;
#if defined(ENABLE_V17)
                case 73:
                    new_modem = (new_transmit)  ?  T31_V17_TX  :  T31_V17_RX;
                    s->short_train = FALSE;
                    s->bit_rate = 7200;
                    break;
                case 74:
                    new_modem = (new_transmit)  ?  T31_V17_TX  :  T31_V17_RX;
                    s->short_train = TRUE;
                    s->bit_rate = 7200;
                    break;
                case 97:
                    new_modem = (new_transmit)  ?  T31_V17_TX  :  T31_V17_RX;
                    s->short_train = FALSE;
                    s->bit_rate = 9600;
                    break;
                case 98:
                    new_modem = (new_transmit)  ?  T31_V17_TX  :  T31_V17_RX;
                    s->short_train = TRUE;
                    s->bit_rate = 9600;
                    break;
                case 121:
                    new_modem = (new_transmit)  ?  T31_V17_TX  :  T31_V17_RX;
                    s->short_train = FALSE;
                    s->bit_rate = 12000;
                    break;
                case 122:
                    new_modem = (new_transmit)  ?  T31_V17_TX  :  T31_V17_RX;
                    s->short_train = TRUE;
                    s->bit_rate = 12000;
                    break;
                case 145:
                    new_modem = (new_transmit)  ?  T31_V17_TX  :  T31_V17_RX;
                    s->short_train = FALSE;
                    s->bit_rate = 14400;
                    break;
                case 146:
                    new_modem = (new_transmit)  ?  T31_V17_TX  :  T31_V17_RX;
                    s->short_train = TRUE;
                    s->bit_rate = 14400;
                    break;
#endif
                default:
                    return FALSE;
                }
                fprintf(stderr, "Short training = %d, bit rate = %d\n", s->short_train, s->bit_rate);
                fprintf(stderr, "Data\n");
                if (s->transmit)
                    s->state = AT_MODE_STUFFED;
                restart_modem(s, new_modem, new_transmit);
                break;
            }
        }
        else
        {
            at_put_response_code(s, RESPONSE_CODE_OK);
        }
    }
    return TRUE;
}
/*- End of function --------------------------------------------------------*/

static void at_command_interpreter(t31_state_t *s, const char *cmd)
{
    uint8_t byte_val;
    int val;
    int r;
    int b;
    const char *t;
    char *u;
    char *v;
    int err = FALSE;
    int ok = TRUE;
    char num[100 + 1];
    char local_party_name[100 + 1];
    int local;
    char ch;
    char buf[100];
    tone_gen_descriptor_t tone_desc;

    /* The spec says the commands within a command line are executed in order, until
       an error is found, or the end of the command line is reached. */
    if (s->echo)
    {
        strcpy(s->rx_data + s->rx_data_bytes, cmd);
        s->rx_data_bytes += strlen(cmd);
    }
fprintf(stderr, "Process command '%s'\n", cmd);
    while (s->state == AT_MODE_COMMAND  &&  !err  &&  *cmd)
    {
        switch (*cmd++)
        {
        case ' ':
        case ';':
            /* Spaces and semicolons are OK - they may delimit commands within a string. */
            break;
        case 'A':
            /* Answer */
            if (s->call_control_handler)
                s->call_control_handler(s, s->call_control_user_data, "");
            /* Send called terminal identification (CED) */
            /* 0.2s of silence, then 2.6s to 4s of 2100Hz tone, then 75ms of
               silence. */
            s->silent_samples += SAMPLE_RATE*0.2;
            make_tone_gen_descriptor(&tone_desc,
                                     2100,
                                     -11,
                                     0,
                                     0,
                                     2600,
                                     75,
                                     0,
                                     0,
                                     FALSE);
            tone_gen_init(&(s->tone_gen), &tone_desc);
            s->modem = T31_CED_TONE;
            s->connection_established = TRUE;
            s->force_fax_mode = TRUE;
            ok = FALSE;
            break;
        case 'D':
            /* Dial */
            ok = FALSE;
            s->force_fax_mode = FALSE;
            local = FALSE;
            /* There are a numbers of options in a dial command string.
               Many are completely irrelevant in this application. */
            t = cmd;
            u = num;
            v = local_party_name;
            for (  ;  (ch = *cmd) != '\r'  &&  !err;  cmd++)
            {
                if (isdigit(ch))
                {
                    if (local)
                        *v++ = ch;
                    else
                        *u++ = ch;
                }
                else
                {
                    switch (ch)
                    {
                    case '.':
                    case ' ':
                    case '-':
                    case 'T':
                    case 'P':
                        break;
                    case 'F':
                        s->force_fax_mode = TRUE;
                        break;
                    case 'V':
                        s->force_fax_mode = FALSE;
                        break;
                    case 'L':
                        v = local_party_name;
                        local = TRUE;
                        break;
                    case 'D':
                        local = FALSE;
                        break;
                    default:
                        err = TRUE;
                        break;
                    }
                }
            }
            *u = '\0';
            *v = '\0';
            if (!err)
            {
                s->state = AT_MODE_DIALING;
                /* TODO: Dial and wait with a 60s timeout */
                if (s->call_control_handler)
                    s->call_control_handler(s, s->call_control_user_data, num);
                switch (at_dial_out(s, t))
                {
                case 1:
                    break;
                case 2:
                    s->force_fax_mode = FALSE;
                    s->state = AT_MODE_COMMAND;
                    at_put_response_code(s, RESPONSE_CODE_BUSY);
                    break;
                case 3:
                    s->force_fax_mode = FALSE;
                    s->state = AT_MODE_COMMAND;
                    at_put_response_code(s, RESPONSE_CODE_NO_DIALTONE);
                    break;
                default:
                    err = TRUE;
                    break;
                }
            }
            /* Send calling tone (CNG) */
            /* 0.5s of 1100Hz + 3.0s of silence repeating */
            make_tone_gen_descriptor(&tone_desc,
                                     1100,
                                     -11,
                                     0,
                                     0,
                                     500,
                                     3000,
                                     0,
                                     0,
                                     TRUE);
            tone_gen_init(&(s->tone_gen), &tone_desc);
            /* Do V.21/HDLC receive in parallel. The other end may send its
               first message at any time. The CNG tone will continue until
               we get a valid preamble. */
            s->hdlc_final = FALSE;
            hdlc_rx_init(&(s->hdlcrx), hdlc_accept, s);
            hdlc_rx_bad_frame_control(&(s->hdlcrx), TRUE);
            restart_modem(s, T31_V21_RX, TRUE);
            s->state = AT_MODE_HDLC;
            s->modem = T31_CNG_TONE;
            break;
        case 'E':
            /* Turn echo on/off */
            if ((val = parse_num(&cmd, 1)) >= 0)
                s->echo = val;
            else
                err = TRUE;
            break;
        case 'H':
            /* On/Off-hook */
            if ((val = parse_num(&cmd, 0)) >= 0)
            {
                /* ATH & ATH0 */
                if (s->call_control_handler)
                    s->call_control_handler(s, s->call_control_user_data, NULL);
                s->connection_established = FALSE;
            }
            else
            {
                err = TRUE;
            }
            break;
        case 'I':
            /* Information */
            switch (val = parse_num(&cmd, 255))
            {
            case 0:
                at_put_response(s, model);
                break;
            case 3:
                at_put_response(s, manufacturer);
                break;
            case 8:
                sprintf(buf, "NMBR = %s", s->originating_number);
                at_put_response(s, buf);
                break;
            case 9:
                sprintf(buf, "NDID = %s", s->destination_number);
                at_put_response(s, buf);
                break;
            default:
                err = TRUE;
                break;
            }
            break;
        case 'L':
            /* Turn Speaker on/off */
            if (parse_num(&cmd, 255) < 0)
                err = TRUE;
            break;
        case 'M':
            /* Turn Speaker on/off */
            if (parse_num(&cmd, 255) < 0)
                err = TRUE;
            break;
        case 'N':
            /* Ring volume */
            if (parse_num(&cmd, 255) < 0)
                err = TRUE;
            break;
        case 'O':
            /* Go online */
            if (parse_num(&cmd, 255) < 0)
                err = TRUE;
            break;
        case 'Q':
            /* Turn result codes on/off */
            switch (val = parse_num(&cmd, 255))
            {
            case 0:
                s->verbose = ASCII_RESULT_CODES;
                break;
            case 1:
                s->verbose = NO_RESULT_CODES;
                break;
            default:
                err = TRUE;
                break;
            }
            break;
        case 'S':
            /* Set/Get Register */
            if ((r = parse_num(&cmd, 99)) >= 0)
            {
                switch (*cmd++)
                {
                case '=':
                    switch (*cmd)
                    {
                    case '?':
                        cmd++;
                        at_put_3digit_numeric_response(s, 0);
                        break;
                    default:
                        val = parse_num(&cmd, 255);
                        if (val < 0  ||  !at_set_reg(s, r, (uint8_t) val))
                            err = TRUE;
                        break;
                    }
                    break;
                case '?':
                    if (at_get_reg(s, r, &byte_val))
                        at_put_3digit_numeric_response(s, (unsigned int) byte_val);
                    else
                        err = TRUE;
                    break;
                case '.':
                    if ((b = parse_num(&cmd, 7)) >= 0)
                    {
                        switch (*cmd++)
                        {
                        case '=':
                            switch (*cmd)
                            {
                            case '?':
                                cmd++;
                                at_put_numeric_response(s, 0);
                                break;
                            default:
                                val = parse_num(&cmd, 1);
                                if (val < 0  ||  !at_set_bit(s, r, b, val))
                                    err = TRUE;
                                break;
                            }
                            break;
                        case '?':
                            if (at_get_bit(s, r, b, &val))
                                at_put_numeric_response(s, (unsigned int) val);
                            else
                                err = TRUE;
                            break;
                        default:
                            err = TRUE;
                            break;
                        }
                    }
                    else
                    {
                        err = TRUE;
                    }
                    break;
                default:
                    err = TRUE;
                    break;
                }
            }
            else
            {
                err = TRUE;
            }
            break;
        case 'V':
            /* Numeric or ASCII result codes */
            switch (val = parse_num(&cmd, 1))
            {
            case 0:
                s->verbose = NUMERIC_RESULT_CODES;
                break;
            case 1:
                s->verbose = ASCII_RESULT_CODES;
                break;
            default:
                err = TRUE;
                break;
            }
            break;
        case 'X':
            /* Result code selection and call progress monitoring control */
            switch (val = parse_num(&cmd, 4))
            {
            case 0:
                break;
            case 1:
                break;
            case 2:
                break;
            case 3:
                break;
            case 4:
                break;
            default:
                err = TRUE;
                break;
            }
            break;
        case 'Z':
            /* Load registers from profile */
            val = parse_num(&cmd, sizeof(profiles)/sizeof(profiles[0]) - 1);
            if (val >= 0)
            {
                if (s->call_control_handler)
                    s->call_control_handler(s, s->call_control_user_data, NULL);
                s->p = profiles[val];
            }
            else 
            {
                err = TRUE;
            }
            break;
        case '+':
            switch (*cmd++)
            {
            case 'F':
                /* Some kind of FAX command */
                switch (*cmd++)
                {
                case 'A':
                    switch (*cmd)
                    {
                    case 'A':
                        /* +FAA */
                        if (!parse_out(s, &cmd, NULL, 255, "0"))
                            err = TRUE;
                        break;
                    case 'R':
                        /* +FAR - adaptive reception control */
                        if (!parse_out(s, &cmd, NULL, 255, "0"))
                            err = TRUE;
                        break;
                    default:
                        err = TRUE;
                        break;
                    }
                    break;
                case 'C':
                    if (strncmp(cmd, "LASS", 4) == 0)
                    {
                        /* +FCLASS */
                        cmd += 4;
                        if (!parse_out(s, &cmd, NULL, 1, "1"))
                            err = TRUE;
                    }
                    else if (strncmp(cmd, "L=", 2) == 0)
                    {
                        /* +FCL - carrier loss timeout */
                        s->carrier_loss_timeout = parse_num(&cmd, 255);
                    }
                    else
                    {
                        err = TRUE;
                    }
                    break;
                case 'D':
                    switch (*cmd)
                    {
                    case 'D':
                        /* +FDD - Double escape character replacement control */
                        s->double_escape = parse_num(&cmd, 255);
                        break;
                    default:
                        err = TRUE;
                        break;
                    }
                    break;
                case 'I':
                    switch (*cmd)
                    {
                    case 'T':
                        /* +FIT - DTE inactivity timeout */
                        s->dte_inactivity_timeout = parse_num(&cmd, 255);
                        break;
                    default:
                        err = TRUE;
                        break;
                    }
                    break;
                case 'L':
                    switch (*cmd)
                    {
                    case 'O':
                        /* +FLO - flow control */
                        break;
                    default:
                        err = TRUE;
                        break;
                    }
                    break;
                case 'M':
                    switch (*cmd)
                    {
                    case 'I':
                        /* +FMI - report manufacturer */
                        at_put_response(s, manufacturer);
                        break;
                    case 'M':
                        /* +FMM - report model ID */
                        at_put_response(s, model);
                        break;
                    case 'R':
                        /* +FMR - report revision ID */
                        at_put_response(s, revision);
                        break;
                    default:
                        err = TRUE;
                        break;
                    }
                    break;
                case 'P':
                    switch (*cmd)
                    {
                    case 'R':
                        break;
                    default:
                        err = TRUE;
                        break;
                    }
                    err = TRUE;
                    break;
                case 'R':
                    switch (*cmd)
                    {
                    case 'H':
                        /* +FRH - receive HDLC frames */
                    case 'M':
                        /* +FRM - receive data */
                    case 'S':
                        /* +FRS - receive silence */
                        cmd++;
                        ok = FALSE;
                        if (!process_class1_cmd(s, &cmd))
                            err = TRUE;
                        break;
                    default:
                        if (strncmp(cmd, "EV", 2) == 0)
                        {
                            /* +FREV */
                            cmd += 2;
                            switch (*cmd++)
                            {
                            case '?':
                                at_put_response(s, revision);
                                break;
                            default:
                                err = TRUE;
                                break;
                            }
                        }
                        else
                        {
                            err = TRUE;
                        }
                        break;
                    }
                    break;
                case 'T':
                    switch (*cmd)
                    {
                    case 'H':
                        /* +FTH - transmit HDLC frames */
                    case 'M':
                        /* +FTM - transmit data */
                    case 'S':
                        /* +FTS - stop transmission and pause */
                        cmd++;
                        ok = FALSE;
                        if (!process_class1_cmd(s, &cmd))
                            err = TRUE;
                        break;
                    default:
                        err = TRUE;
                        break;
                    }
                    break;
                default:
                    err = TRUE;
                    break;
                }
                break;
            case 'G':
                switch (*cmd++)
                {
                case 'M':
                    switch (*cmd)
                    {
                    case 'I':
                        /* +GMI */
                        at_put_response(s, manufacturer);
                        break;
                    case 'M':
                        /* +GMM */
                        at_put_response(s, model);
                        break;
                    case 'R':
                        /* +GMR */
                        at_put_response(s, revision);
                        break;
                    default:
                        err = TRUE;
                        break;
                    }
                    break;
                default:
                    err = TRUE;
                    break;
                }
                break;
            case 'I':
                switch (*cmd++)
                {
                case 'P':
                    switch (*cmd)
                    {
                    case 'C':
                        /* +IPC */
                        at_put_response(s, manufacturer);
                        break;
                    case 'R':
                        /* +IPR */
                        at_put_response(s, model);
                        break;
                    default:
                        err = TRUE;
                        break;
                    }
                    break;
                default:
                    err = TRUE;
                    break;
                }
                break;
            case 'V':
                /* Voice */
                break;
            default:
                err = TRUE;
                break;
            }
            break;
        case '&':
            switch (*cmd++)
            {
            case 'C':
                /* &C */
                if ((val = parse_num(&cmd, 1)) >= 0)
                {
                }
                else
                {
                    err = TRUE;
                }
                break;
            case 'D':
                /* &D */
                if ((val = parse_num(&cmd, 3)) >= 0)
                {
                }
                else
                {
                    err = TRUE;
                }
                break;
            case 'F':
                /* &F */
                if (s->call_control_handler)
                    s->call_control_handler(s, s->call_control_user_data, NULL);
                s->p = profiles[0];
                break;
            case 'H':
                /* &H */
                if ((val = parse_num(&cmd, 7)) >= 0)
                    at_set_bits(s, 27, 3, 5, (uint8_t) val);
                else
                    err = TRUE;
                break;
            default:
                err = TRUE;
                break;
            }
            break;
        case '#':
            if (strncmp(cmd, "CID", 3) == 0)
            {
                /* #CID */
                val = at_cid_mode(s, -1);
                if (!parse_out(s, &cmd, &val, 10, "0,10"))
                {
                   err = TRUE;
                }
                else
                {
                    switch (val)
                    {
                    case 0:
                    case 10:
                        at_cid_mode(s, val);
                        break;
                    default:
                        err = TRUE;
                        break;
                    }
                }
            }
            else
            {
                err = TRUE;
            }
            break;
        default:
            err = TRUE;
            break;
        }
    }
    if (err)
        at_put_response_code(s, RESPONSE_CODE_ERROR);
    else if (ok)
        at_put_response_code(s, RESPONSE_CODE_OK);
}
/*- End of function --------------------------------------------------------*/

void at_interpreter(t31_state_t *s, const char *cmd, int len)
{
    int i;
    int c;

    for (i = 0;  i < len;  i++)
    {
        /* The spec says the top bit should be ignored */
        c = *cmd++ & 0x7F;
        /* Handle incoming character */
        if (s->line_ptr < 2)
        {
            /* Look for the initial AT, and ignore anything before it */
            if (s->line_ptr == 0)
            {
                if (tolower(c) == 'a')
                    s->line[s->line_ptr++] = c;
            }
            else
            {
                if (tolower(c) == 't')
                {
                    s->line[s->line_ptr++] = c;
                }
                else
                {
                    s->line_ptr = 0;
                    if (tolower(c) == 'a')
                        s->line[s->line_ptr++] = c;
                }
            }
        }
        else
        {
            /* We are beyond the initial AT */
            if (c >= 0x20)
            {
                /* Add a new char */
                if (s->line_ptr < (sizeof(s->line) - 1))
                    s->line[s->line_ptr++] = toupper(c);
            }
            else
            {
                switch (c)
                {
                case '\n':
                case '\r':
                    /* Line validation */
                    s->line[s->line_ptr] = '\0';
                    if (s->line_ptr > 2)
                        at_command_interpreter(s, s->line);
                    s->line_ptr = 0;
                    break;
                case '\b':
                    /* Backspace */
                    if (s->line_ptr > 0)
                        s->line_ptr--;
                    break;
                default:
                    /* The spec says control characters, other than those
                       explicitly handled, should be ignored. */
                    break;
                }
            }
        }
    }
}
/*- End of function --------------------------------------------------------*/

void t31_at_rx(t31_state_t *s, const char *t, int len)
{
    if (s->state == AT_MODE_COMMAND)
    {
        at_interpreter(s, t, len);
    }
    else if (s->state == AT_MODE_CONNECTED)
    {
    }
    else if (s->state == AT_MODE_HDLC)
    {
        dle_unstuff_hdlc(s, t, len);
    }
    else if (s->state == AT_MODE_STUFFED)
    {
        dle_unstuff(s, t, len);
    }
}
/*- End of function --------------------------------------------------------*/

int t31_rx(t31_state_t *s, int16_t *buf, int len)
{
    int i;
    int read_len;

    if ((read_len = read(s->fd, s->tx_data, sizeof(s->tx_data))) > 0)
    {
#if 0
        for (i = 0;  i < read_len;  i++)
            fprintf(stderr, "%c", s->tx_data[i]);
#endif
        t31_at_rx(s, s->tx_data, read_len);
    }
    if (!s->transmit  ||  s->modem == T31_CNG_TONE)
    {
        switch (s->modem)
        {
        case T31_CED_TONE:
            break;
        case T31_CNG_TONE:
        case T31_V21_RX:
            fsk_rx(&(s->v21rx), buf, len);
            break;
#if defined(ENABLE_V17)
        case T31_V17_RX:
            v17_rx(&(s->v17rx), buf, len);
            break;
#endif
        case T31_V27TER_RX:
            v27ter_rx(&(s->v27ter_rx), buf, len);
            break;
        case T31_V29_RX:
            v29_rx(&(s->v29rx), buf, len);
            break;
        default:
            /* Absorb the data, but ignore it. */
            break;
        }
    }
    if (s->rx_data_bytes)
    {
#if 0
        fprintf(stderr, "reply %d\n", s->rx_data_bytes);
        for (i = 0;  i < s->rx_data_bytes;  i++)
            fprintf(stderr, "%c", s->rx_data[i]);
        fprintf(stderr, "\n");
#endif
        write(s->fd, s->rx_data, s->rx_data_bytes);
        s->rx_data_bytes = 0;
    }
    return  0;
}
/*- End of function --------------------------------------------------------*/

int t31_tx(t31_state_t *s, int16_t *buf, int max_len)
{
    int len;
    int lenx;

    len = 0;
    if (s->transmit)
    {
        if (s->silent_samples)
        {
            len = s->silent_samples;
            if (len > max_len)
                len = max_len;
            s->silent_samples -= len;
            max_len -= len;
            memset(buf, 0, len*sizeof(int16_t));
            if (s->silent_samples <= 0)
                at_put_response_code(s, RESPONSE_CODE_OK);
        }
        if (max_len > 0)
        {
            switch (s->modem)
            {
            case T31_CED_TONE:
                lenx = tone_gen(&(s->tone_gen), buf + len, max_len);
                len += lenx;
                if (lenx <= 0)
                {
                    /* Go directly to V.21/HDLC transmit. */
                    s->hdlc_final = FALSE;
                    hdlc_tx_init(&(s->hdlctx), hdlc_tx_underflow, s);
                    hdlc_tx_preamble(&(s->hdlctx), 40);
                    restart_modem(s, T31_V21_TX, TRUE);
                    s->state = AT_MODE_HDLC;
                    at_put_response_code(s, RESPONSE_CODE_OK);
                }
                break;
            case T31_CNG_TONE:
                len += tone_gen(&(s->tone_gen), buf + len, max_len);
                break;
            case T31_V21_TX:
                len += fsk_tx(&(s->v21tx), buf + len, max_len);
                break;
#if defined(ENABLE_V17)
            case T31_V17_TX:
                len += v17_tx(&(s->v17tx), buf + len, max_len);
                break;
#endif
            case T31_V27TER_TX:
                len += v27ter_tx(&(s->v27ter_tx), buf + len, max_len);
                break;
            case T31_V29_TX:
                len += v29_tx(&(s->v29tx), buf + len, max_len);
                break;
            }
        }
    }
    return len;
}
/*- End of function --------------------------------------------------------*/

void t31_ring(t31_state_t *s)
{
    at_put_response_code(s, RESPONSE_CODE_RING);
}
/*- End of function --------------------------------------------------------*/

int t31_init(t31_state_t *s,
             const char *dev,
             t31_call_control_handler_t *call_control_handler,
             void *user_data)
{
    int fd;
    
    fd = open(dev, O_RDWR | O_NDELAY);
    if (fd < 0)
        return -1;
    memset(s, 0, sizeof(*s));
    s->fd = fd;
#if defined(ENABLE_V17)
    v17_rx_init(&(s->v17rx), s->bit_rate, fast_putbit, s);
    v17_tx_init(&(s->v17tx), s->bit_rate, fast_getbit, s);
#endif
    v27ter_rx_init(&(s->v27ter_rx), s->bit_rate, fast_putbit, s);
    v27ter_tx_init(&(s->v27ter_tx), s->bit_rate, fast_getbit, s);
    v29_rx_init(&(s->v29rx), s->bit_rate, fast_putbit, s);
    v29_tx_init(&(s->v29tx), s->bit_rate, fast_getbit, s);
    s->rx_signal_present = FALSE;

    s->line_ptr = 0;
    s->state = AT_MODE_COMMAND;
    s->connection_established = FALSE;
    s->force_fax_mode = FALSE;
    s->originating_number[0] = '\0';
    s->destination_number[0] = '\0';
    s->modem = -1;
    s->transmit = -1;
    s->verbose = TRUE;
    s->echo = TRUE;
    if (queue_create(&(s->rx_queue), 4096, QUEUE_WRITE_ATOMIC | QUEUE_READ_ATOMIC) < 0)
    {
        close(fd);
        return -1;
    }
    s->call_control_handler = call_control_handler;
    s->call_control_user_data = user_data;
    return 0;
}
/*- End of function --------------------------------------------------------*/
/*- End of file ------------------------------------------------------------*/
